# -*- coding: utf-8 -*-
import re
from typing import List

from django.conf import settings
from drf_spectacular.contrib.rest_framework_simplejwt import SimpleJWTScheme
from drf_spectacular.drainage import warn
from drf_spectacular.generators import SchemaGenerator
from drf_spectacular.openapi import AutoSchema
from drf_spectacular.plumbing import build_bearer_security_scheme_object, get_doc
from rest_framework import serializers
from rest_framework.request import Request


def get_action_doc(action_or_method, docstring, actions_sep):
    action = action_or_method.__name__ if not isinstance(action_or_method, str) else action_or_method

    if action in docstring:
        try:
            search = re.search(rf"{actions_sep} {action}:[\s\n](.*)\n", docstring)
            title = search.groups()[0].strip()
        except AttributeError:
            # No match on title, skip instance
            return None
        try:
            description = docstring[search.end():].split(actions_sep)[0]
        except AttributeError:
            warn(f'Cannot detect API description for action {action}. Setting description as blank string.')
            description = ''
        return f"## {title}\n### {description}"
    return None


class ThuxAutoSchema(AutoSchema):

    exclude_patterns = settings.THX_OPENAPI_TAGS_EXCLUDE_PATTERNS
    response_serializer_meta_field = "response_serializer"

    def get_response_serializers(self):
        response_serializers = super().get_response_serializers()
        if hasattr(response_serializers.__class__, "Meta"):
            response_serializers = getattr(
                response_serializers.__class__.Meta,
                self.response_serializer_meta_field,
                response_serializers
            )
        return response_serializers

    def _get_serializer(self):
        # PATCH that handle serializers profiles if standard SchemaGenerator is used.
        # Use by default the `superuser` profile.
        self.view.kwargs["profile"] = self.view.kwargs.get("profile", settings.THX_OPENAPI_DEFAULT_PROFILE)

        try:
            self.view.get_serializer_class()
        except AssertionError:
            class OpenApiSerializer(serializers.Serializer):
                # PATCH that solve warning: Component name "" contains illegal characters.
                # Only "A-Z a-z 0-9 - . _" are allowed.
                # Furthermore, "-" and "." are discoursed due to potential tooling issues.
                # This likely leads to an invalid schema.
                pass
            self.view.serializer_class = OpenApiSerializer

        # TODO valutare se serve gestire l'assenza del queryset
        # try:
        #     self.view.get_queryset()
        # except Exception:
        #     # PATCH that handle custom views were no queryset is defined
        #     self.view.queryset = "-"

        return super()._get_serializer()

    def get_tags(self) -> List[str]:
        """
        Exclude some path regex patterns from the allowed tags.
        Default is exclude pattern will exclude from valid paths all v1, v2, v3, ... paths.
        """
        indexes = [0, 1]
        tokenized_path = self._tokenize_path()

        do = True
        while do:
            do, indexes = self._get_tag_indexes(tokenized_path, indexes)

        return tokenized_path[indexes[0]:indexes[1]]

    def _get_tag_indexes(self, tokenized_path, indexes):
        for pattern in self.exclude_patterns:
            if re.match(pattern, tokenized_path[indexes[0]]):
                indexes = [i + 1 for i in indexes]
                return True, indexes
        return False, indexes

    def get_description(self):
        """Provide action docs parsed from ViewSet docstring."""
        action_or_method = getattr(self.view, getattr(self.view, 'action', self.method.lower()), None)
        view_doc = get_doc(self.view.__class__)
        view_action_doc = get_action_doc(action_or_method, view_doc, settings.THX_OPENAPI_DOCSTRING_ACTIONS_SEPARATOR)
        action_doc = get_doc(action_or_method)
        return action_doc or view_action_doc or view_doc

    def get_operation_id(self):
        """Get operation ID using path tokenization keeping path variables for the operation ID computing"""
        tokenized_path = self._tokenize_path_with_variables()
        # replace dashes as they can be problematic later in code generation
        tokenized_path = [t.replace('-', '_') for t in tokenized_path]

        if self.method == 'GET' and self._is_list_view():
            action = 'list'
        else:
            action = self.method_mapping[self.method.lower()]

        if not tokenized_path:
            tokenized_path.append('root')

        if re.search(r'<drf_format_suffix\w*:\w+>', self.path_regex):
            tokenized_path.append('formatted')

        return '_'.join(tokenized_path + [action])

    def _tokenize_path_with_variables(self):
        # remove path variables
        path = re.sub(
            pattern=self.path_prefix,
            repl='',
            string=self.path,
            flags=re.IGNORECASE
        )
        # cleanup and tokenize remaining parts.
        path = path.rstrip('/').lstrip('/').replace("{", "").replace("}", "").split('/')
        return [t for t in path if t]


class ThuxSimpleJWTScheme(SimpleJWTScheme):
    priority = 1

    def get_security_definition(self, auto_schema):
        # PATCH that suppress the warning caused by Thux custom setting:
        #   SIMPLE_JWT['AUTH_HEADER_TYPES'] = ('JWT', 'Token',)
        from rest_framework_simplejwt.settings import api_settings

        return build_bearer_security_scheme_object(
            header_name=getattr(api_settings, 'AUTH_HEADER_NAME', 'HTTP_AUTHORIZATION'),
            token_prefix=api_settings.AUTH_HEADER_TYPES[0],
            bearer_format='JWT'
        )


class ThuxSchemaGenerator(SchemaGenerator):
    """
    Schema Generator that injects the url param `profile` or use the default value configured
    in THX_OPENAPI_DEFAULT_PROFILE setting
    """
    input_request: Request

    def parse(self, input_request, public):
        self.input_request = input_request
        return super().parse(input_request, public)

    def _get_paths_and_endpoints(self):
        endpoints = super()._get_paths_and_endpoints()
        try:
            profile = self.input_request.query_params.get("profile", settings.THX_OPENAPI_DEFAULT_PROFILE)
        except AttributeError:
            profile = settings.THX_OPENAPI_DEFAULT_PROFILE
        replace_profile_in_urls = settings.THX_OPENAPI_REPLACE_PROFILE_IN_URLS
        for index, (path, path_regex, method, view) in enumerate(endpoints):
            view.kwargs = {"profile": profile}
            if replace_profile_in_urls:
                path = path.replace("{profile}", profile)
                endpoints[index] = (path, path_regex, method, view)
        return endpoints
